# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_atyp.ipynb.

# %% auto 0
__all__ = ['P', 'R', 'Operator', 'Shaped', 'DataType', 'GraphType']

# %% ../nbs/01_atyp.ipynb 4
from types import BuiltinFunctionType
from typing import Callable, Union, TypeAlias, ParamSpec, TypeVar, TypeGuard

# %% ../nbs/01_atyp.ipynb 5
from .cons import ShapedTypes, DataTypes, GraphTypes

# %% ../nbs/01_atyp.ipynb 7
P = ParamSpec('P')
R = TypeVar('R', bound=Callable[..., TypeGuard[object]])

# %% ../nbs/01_atyp.ipynb 8
Operator: TypeAlias = Union[BuiltinFunctionType, Callable]

# %% ../nbs/01_atyp.ipynb 9
Shaped: TypeAlias = Union[*ShapedTypes, *ShapedTypes]
DataType: TypeAlias = Union[*DataTypes, *DataTypes]
GraphType: TypeAlias = Union[*GraphTypes, *GraphTypes]



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/getr',
                'doc_host': 'https://dsm-72.github.io',
                'git_url': 'https://github.com/dsm-72/getr',
                'lib_path': 'getr'},
  'syms': { 'getr.attr': { 'getr.attr.getaffinity': ('attr.html#getaffinity', 'getr/attr.py'),
                           'getr.attr.getcat': ('attr.html#getcat', 'getr/attr.py'),
                           'getr.attr.getcategories': ('attr.html#getcategories', 'getr/attr.py'),
                           'getr.attr.getcentroids': ('attr.html#getcentroids', 'getr/attr.py'),
                           'getr.attr.getclusters': ('attr.html#getclusters', 'getr/attr.py'),
                           'getr.attr.getcodes': ('attr.html#getcodes', 'getr/attr.py'),
                           'getr.attr.getcols': ('attr.html#getcols', 'getr/attr.py'),
                           'getr.attr.getcolumns': ('attr.html#getcolumns', 'getr/attr.py'),
                           'getr.attr.getdata': ('attr.html#getdata', 'getr/attr.py'),
                           'getr.attr.getdf': ('attr.html#getdf', 'getr/attr.py'),
                           'getr.attr.getdfds': ('attr.html#getdfds', 'getr/attr.py'),
                           'getr.attr.getdiffop': ('attr.html#getdiffop', 'getr/attr.py'),
                           'getr.attr.getdiffpot': ('attr.html#getdiffpot', 'getr/attr.py'),
                           'getr.attr.getds': ('attr.html#getds', 'getr/attr.py'),
                           'getr.attr.getembedding': ('attr.html#getembedding', 'getr/attr.py'),
                           'getr.attr.getgraph': ('attr.html#getgraph', 'getr/attr.py'),
                           'getr.attr.geticat': ('attr.html#geticat', 'getr/attr.py'),
                           'getr.attr.getidx': ('attr.html#getidx', 'getr/attr.py'),
                           'getr.attr.getiloc': ('attr.html#getiloc', 'getr/attr.py'),
                           'getr.attr.getindex': ('attr.html#getindex', 'getr/attr.py'),
                           'getr.attr.getkdegree': ('attr.html#getkdegree', 'getr/attr.py'),
                           'getr.attr.getkernel': ('attr.html#getkernel', 'getr/attr.py'),
                           'getr.attr.getloc': ('attr.html#getloc', 'getr/attr.py'),
                           'getr.attr.getncol': ('attr.html#getncol', 'getr/attr.py'),
                           'getr.attr.getnrow': ('attr.html#getnrow', 'getr/attr.py'),
                           'getr.attr.getshape': ('attr.html#getshape', 'getr/attr.py'),
                           'getr.attr.iscatacc': ('attr.html#iscatacc', 'getr/attr.py')},
            'getr.atyp': {},
            'getr.cons': {},
            'getr.core': { 'getr.core._get': ('core.html#_get', 'getr/core.py'),
                           'getr.core.getr': ('core.html#getr', 'getr/core.py'),
                           'getr.core.make': ('core.html#make', 'getr/core.py'),
                           'getr.core.name': ('core.html#name', 'getr/core.py'),
                           'getr.core.onto': ('core.html#onto', 'getr/core.py'),
                           'getr.core.qual': ('core.html#qual', 'getr/core.py')},
            'getr.grds': { 'getr.grds.isdata': ('grds.html#isdata', 'getr/grds.py'),
                           'getr.grds.isgraph': ('grds.html#isgraph', 'getr/grds.py'),
                           'getr.grds.isshaped': ('grds.html#isshaped', 'getr/grds.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_grds.ipynb.

# %% auto 0
__all__ = ['isdata', 'isgraph', 'isshaped']

# %% ../nbs/01_grds.ipynb 4
from typing import TypeGuard
from typing import TypeAlias, Callable, Union

# %% ../nbs/01_grds.ipynb 5
from chck import isof, isdf, isds
from hasr import hasr, hasdf, hasshape
from nlit import SHAPE

# %% ../nbs/01_grds.ipynb 6
from itry import pddataframe, ptdataset

# %% ../nbs/01_grds.ipynb 7
from .atyp import DataType, GraphType, Shaped
from .cons import DataTypes, GraphTypes, ShapedTypes

# %% ../nbs/01_grds.ipynb 10
def isdata(x) -> TypeGuard[DataType]:
    "Returns `True` if `x` is a `DataType`"
    return isof(x, DataTypes)

def isgraph(x) -> TypeGuard[GraphType]:
    "Returns `True` if `x` is a `GraphType`"
    return isof(x, GraphTypes)

# %% ../nbs/01_grds.ipynb 11
def isshaped(x: object) -> TypeGuard[Shaped]:
    return isof(x, ShapedTypes) and hasshape(x)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_core.ipynb.

# %% auto 0
__all__ = ['getr', 'qual', 'name', 'make', 'onto']

# %% ../nbs/02_core.ipynb 4
from functools import wraps
from typing import TypeGuard
from typing import Callable, Optional

# %% ../nbs/02_core.ipynb 5
from chck import iscall, isnone, notnone, isiter
from nlit import GET, __DOC__, __NAME__, __QUALNAME__, __MODULE__, __ANNOTATIONS__

# %% ../nbs/02_core.ipynb 6
from .atyp import R

# %% ../nbs/02_core.ipynb 8
def _get(
    obj: object, val: str, default = None, 
    check: bool = True, guard: Optional[Callable] = notnone, retself: bool = False
):
    if all((check, iscall(guard))) and guard(val): return val
    if all((retself, check, iscall(guard))) and guard(obj): return obj
    return default

# %% ../nbs/02_core.ipynb 9
def getr(
    obj: object, attr: str, 
    check: bool = True, guard: Optional[Callable] = notnone, 
    default = None, otype: Optional[Callable] = None, retself: bool = False
):
    '''Get a specified attribute from an object with an optional value check before defaulting.

    This function retrievesan object's attribute with the option to validate 
    the attribute's value using a guard function. The guard function can be any callable 
    that takes a single argument and returns a boolean. If the attribute is not present, 
    a default value can be considered for the guard check.

    Parameters
    ----------
    obj : object
        The object to check for the attribute.

    attr : str
        The name of the attribute to check for.

    check : bool, optional
        If True, the function performs a guard check on the attribute's value, by default True.

    guard : Optional[Callable], default: `notnone`.
        The guard function used to validate the attribute's value, by default notnone.

    default : Any, optional
        The default value to use if the attribute is not found, by default None.

    otype : Optional[Callable], optional
        The type of the object to check for, by default None.
        If specified, the function will check if the object is of the specified type 
        before checking for the attribute. If this fails, the function will return `default`.

    retself : bool, optional
        If True, the function will return the object if the guard check passes, by default False.

    Returns
    -------
    attribute : Any
        Returns the attribute if the object has `attr` (i.e. `obj.attr` and passes the guard check if `check` is True),
            otherwise `default`.
        

    See Also
    --------
    chck.notnone : A guard function that checks if a value is not None i.e. `val is not None`.
    '''
    if all((check, notnone(otype), iscall(otype))) and not otype(obj): return default
    try: 
        val = getattr(obj, attr, default)
        if isnone(val): val = default        
    except: val = default
    return _get(obj, val, default, check, guard, retself)
    if all((check, iscall(guard))) and guard(val): return val
    if all((retself, check, iscall(guard))) and guard(obj): return obj
    return default # if not check else fallback

# %% ../nbs/02_core.ipynb 10
def qual(attr: str, base: str = GET, prefix: str = '', suffix: str = '') -> str:
    '''Produce a '__qualname__' for a function based on an attribute name.'''
    if isnone(prefix): prefix = ''
    if isnone(suffix): suffix = ''
    return f'{prefix}{base}{attr}{suffix}'
    
def name(fn: Callable, attr: str, base: str = GET, prefix: str = None, suffix: str = None) -> Callable:
    for a in (__NAME__, __QUALNAME__): setattr(fn, a, qual(attr, base, prefix, suffix))
    return fn

# %% ../nbs/02_core.ipynb 11
def make(
    attr: str, check: bool = True, guard: Optional[Callable] = notnone, default = None, 
    otype: Optional[Callable] = None, retself: bool = False
) -> Callable[[R], R]:
    '''Create a decorator for generating attribute-checking functions.

    This decorator factory returns a decorator that, when applied to a function,
    creates a new function which retrieves a specified attribute from object has a specified attribute (named `attr`).
    Additionally, it can validate the attribute's value using a guard function `guard`.
    The decorator updates the original function's docstring to include the new parameters.

    Parameters
    ----------
    attr : str
        The name of the attribute to check for in the decorated function.

    check : bool, default: True
        If True, the decorated function performs a guard check on the attribute's value.

    guard : Optional[Callable], default: notnone
        The guard function used to validate the attribute's value in the decorated function. 
        Defaults to `notnone`, which checks that the value is not None.

    default : Any, optional
        The default value to use if the attribute is not found in the decorated function.

    otype : Optional[Callable], optional
        The type of the object to check for, by default None.
        If specified, the function will check if the object is of the specified type 
        before checking for the attribute. If this fails, the function will return `default`.

    retself : bool, optional
        If True, the function will return the object if the guard check passes, by default False.

    Returns
    -------
    Callable[[R], R]
        A decorator that, when applied to a function, creates a new attribute-retrieving function.
        This new function includes additional parameters and updated docstring.    
    '''
    def deco(fn: Callable[..., TypeGuard[object]]):
        fndoc = getattr(fn, __DOC__, None)
        if isnone(fndoc): fndoc = f'Get {attr!r} i.e. `obj.{attr}` from an object.'

        doc = f'''
            {fndoc}

            Parameters
            ----------
            obj : object
                The object to get {attr!r} from i.e. `obj.{attr}`.
                
            check : bool, default: {check}
                If True, the function performs a typeguard check on the `obj.{attr}` value.
        
            guard : Optional[Callable], default: `notnone`.
                The guard function used to validate `obj.{attr}`'s value if `check == True`.
        
            default : Any, optional
                The default value to use if `obj` does not have {attr!r}.

            otype : Optional[Callable], default: {otype}
                The type of the object to check for, by default None.
                If specified, the function will check if the object is of the specified type 
                before checking for the attribute. If this fails, the function will return `default`.

            retself : bool, default: {retself}
                If True, the function will return the object if the guard check passes, by default False.
        
            Returns
            -------
            attribute : Any
                Returns the attribute if the object has {attr!r} (i.e. `obj.{attr}` and passes the guard check if `check` is True),
                otherwise `{default}`.
        
            See Also
            --------
            chck.notnone : A guard function that checks if a value is not None i.e. `val is not None`.
            '''
        
        @wraps(fn, assigned=(__MODULE__, __NAME__, __QUALNAME__, ))
        def wrap(
            obj: object, check: bool = check, guard: Optional[Callable] = guard, default = default, otype = otype,
            retself: bool = retself
        ) -> TypeGuard[object]:
            return getr(obj, attr, check=check, guard=guard, default=default, otype=otype, retself=retself)
        wrap.__doc__ = doc
        wrap = name(wrap, attr)        
        return wrap
    return deco

# %% ../nbs/02_core.ipynb 12
def onto(
    getter,    
    check = True, 
    guard: Optional[Callable] = notnone, 
    default = None, 
    otype: Optional[Callable] = None, 
    retself: bool = False
): 
    def deco(fn: Callable[..., TypeGuard[object]]):
        @wraps(fn, assigned=(__MODULE__, __NAME__, __QUALNAME__, ))
        def wrap(obj: object, *args, **kwargs) -> TypeGuard[object]:
            val = getter(obj, *args, **kwargs)
            if all((check, notnone(otype), iscall(otype))) and not otype(val): 
                val = obj
            res = fn(val, *args, **kwargs)
            res = _get(val, res, default, check, guard, retself)
            eqs = res == default
            alltrue = isiter(eqs) and all(eqs)
            onetrue = not isiter(eqs) and eqs
            if (alltrue) or (onetrue):
            # if res == default:
                res = fn(obj, *args, **kwargs)
                res = _get(obj, res, default, check, guard, retself)
            return res
        return wrap
    return deco



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_cons.ipynb.

# %% auto 0
__all__ = ['DataTypes', 'ShapedTypes', 'GraphTypes', 'empty']

# %% ../nbs/00_cons.ipynb 4
from typing import TypeGuard, Union

# %% ../nbs/00_cons.ipynb 6
from itry import npmatrix, nparray, sparray, spmatrix, anndata
from itry import bgraph, dgraph, kgraph, tgraph, lgraph, mgraph, tlgraph, klgraph, mlgraph
from itry import pdindex, pdseries, pdcatindex, pddataframe
from itry import pttensor, ptinttensor, ptbooltensor, ptfloattensor
from itry import earray

# %% ../nbs/00_cons.ipynb 7
DataTypes = (npmatrix, nparray, sparray, spmatrix, anndata)
ShapedTypes = (*DataTypes, pdindex, pdseries, pdcatindex, pddataframe, pttensor, ptinttensor, ptbooltensor, ptfloattensor)
GraphTypes = (bgraph, dgraph, kgraph, tgraph, lgraph, mgraph, tlgraph, klgraph, mlgraph)

empty = earray



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/10_attr.ipynb.

# %% auto 0
__all__ = ['getshape', 'getnrow', 'getncol', 'getdf', 'getcols', 'getcolumns', 'iscatacc', 'getcat', 'getcodes', 'getcategories',
           'getidx', 'getindex', 'getloc', 'getiloc', 'geticat', 'getds', 'getdfds', 'getkernel', 'getkdegree',
           'getaffinity', 'getdiffop', 'getdata', 'getgraph', 'getclusters', 'getembedding', 'getcentroids',
           'getdiffpot']

# %% ../nbs/10_attr.ipynb 4
from typing import TypeGuard
from typing import Callable

# %% ../nbs/10_attr.ipynb 5
from itry import PHATE, earray
from nlit import (
    KERNEL, KERNEL_DEGREE, AFFINITY, DIFF_OP,  GRAPH, CLUSTERS, EMBEDDING,
    CENTROIDS, DIFF_POTENTIAL, CENTROIDS, DIFF_OP, AFFINITY, CLUSTERS, GRAPH,
    DATA, SHAPE, INDEX, COLUMNS, DF, LOC, ILOC, ICAT, IDX, COLS, DS, ROWS,
    CAT, CODES, CATEGORY
)

from chck import (
    isdf, isds, isseries, isdfds, isint, istup, iscatidx, isidxlike, 
    isdelegate, iscataccessor, anyguard, isidx
)
from hasr import has1, has2
from hasr import hascat, hascodes, hasastype, hascategories

# %% ../nbs/10_attr.ipynb 6
from .cons import empty
from .atyp import DataType, GraphType, Shaped
from .grds import isdata, isgraph, isshaped
from .core import make, getr, onto

# %% ../nbs/10_attr.ipynb 9
@make(SHAPE, guard=istup, otype=isshaped, default=tuple(), retself=True)
def getshape(x, default = tuple()) -> tuple | None: ...

# %% ../nbs/10_attr.ipynb 11
@onto(getshape, guard=isint, retself=True, default=0, otype=has1)
def getnrow(shape: tuple, default = 0) -> tuple | None: 
    print(shape)
    if len(shape) > 0 and isint(shape[0]): return shape[0]
    try: return shape[0]
    except: return default

@onto(getshape, guard=isint, retself=True, default=0, otype=has2)
def getncol(shape: tuple, default = 0) -> tuple | None:    
    if len(shape) > 1 and isint(shape[1]): return shape[1]
    try: return shape[1]
    except: return default

# %% ../nbs/10_attr.ipynb 14
@make(DF, guard=isdf, retself=True)
def getdf(x, default = None) -> DataType | None: ...

@make(COLS, otype=isshaped)
def getcols(x) -> DataType | None: ...

@make(COLUMNS, otype=isshaped)
def getcolumns(x) -> DataType | None: ...

# %% ../nbs/10_attr.ipynb 15
def iscatacc(x) -> bool:
    return anyguard(x, iscataccessor, isseries, isidxlike, isdelegate)

@make(CAT, otype=iscatacc, retself=True)
def getcat(x) -> DataType | None: ...

# %% ../nbs/10_attr.ipynb 16
@onto(getcat, guard=iscatacc, retself=True)
def getcodes(series: tuple, default = 0) -> tuple | None:
    if isseries(series) and hasastype(series): series = series.astype(CATEGORY)
    if hascat(series) and hascodes(series.cat): return series.cat.codes
    if iscatidx(series) and hascodes(series): return series.codes
    if iscataccessor(series) and hascodes(series): return series.codes
    return default

# %% ../nbs/10_attr.ipynb 17
@onto(getcat, guard=isidx, retself=True)
def getcategories(series: tuple, default = 0) -> tuple | None:
    if isseries(series) and hasastype(series): series = series.astype(CATEGORY)
    if hascat(series) and hascategories(series.cat): return series.cat.categories
    if iscatidx(series) and hascategories(series): return series.categories
    if iscataccessor(series) and hascategories(series): return series.codes
    return default

# %% ../nbs/10_attr.ipynb 19
@make(IDX)
def getidx(x, default = None) -> DataType | None: ...

@make(INDEX)
def getindex(x, default = None) -> DataType | None: ...

# %% ../nbs/10_attr.ipynb 20
@make(LOC, otype=isdf)
def getloc(x, default = None) -> DataType | None: ...

@make(ILOC, otype=isdf)
def getiloc(x, default = None) -> DataType | None: ...

@make(ICAT, otype=isdf)
def geticat(x, default = None) -> DataType | None: ...

# %% ../nbs/10_attr.ipynb 22
@make(DS, guard=isds, retself=True)
def getds(x, default = None) -> DataType | None: ...

# %% ../nbs/10_attr.ipynb 23
@make(DF, guard=isdf, otype=isdfds, retself=True) 
def getdfds(x, default = None) -> DataType | None: ...

# %% ../nbs/10_attr.ipynb 25
@make(KERNEL, guard=isdata)
def getkernel(x, default = None) -> DataType | None: ...

@make(KERNEL_DEGREE, guard=isdata)
def getkdegree(x, default =  None) -> DataType | None: ...

@make(AFFINITY, guard=isdata)
def getaffinity(x, default =  None) -> DataType | None: ...

@make(DIFF_OP, guard=isdata)
def getdiffop(x, default =  None) -> DataType | None: ...

@make(DATA, guard=isdata, default=earray)
def getdata(x, default =  None) -> DataType | None: ...

@make(GRAPH, guard=isgraph)
def getgraph(p: PHATE, default =  None) -> DataType | None: ...

@make(CLUSTERS, guard=isdata)
def getclusters(x, default =  None) -> DataType | None: ...

@make(EMBEDDING, guard=isdata)
def getembedding(x, default =  None) -> DataType | None: ...

@make(CENTROIDS,guard=isdata)
def getcentroids(x, default =  None) -> DataType | None: ...

@make(DIFF_POTENTIAL, guard=isdata)
def getdiffpot(p: PHATE, default =  None) -> DataType | None: ...
